@@ -1,6 +1,6 @@
 /* grep.c - main driver file for grep.
    Copyright (C) 1992, 1997, 1998, 1999, 2000, 2001, 2002, 2004,
-   2005 Free Software Foundation, Inc.
+   2005, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1506,7 +1506,12 @@ Direct invocation as either `egrep' or `fgrep' is deprecated.\n"));
 With no FILE, or when FILE is -, read standard input.  If less than two FILEs\n\
 are given, assume -h.  Exit status is 0 if any line was selected, 1 otherwise;\n\
 if any error occurs and -q was not given, the exit status is 2.\n"));
-      printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
+      printf (_("\nReport bugs to: %s\n"), PACKAGE_BUGREPORT);
+      printf (_("GNU Grep home page: <%s>\n"),
+              "http://www.gnu.org/software/grep/");
+      fputs (_("General help using GNU software: <http://www.gnu.org/gethelp/>\n"),
+             stdout);
+              
     }
   exit (status);
 }
@@ -1707,18 +1712,20 @@ parse_grep_colors (void)
 	    break;
 	/* If name unknown, go on for forward compatibility.  */
 	if (cap->name)
-	  if (cap->var)
-	    {
-	      if (val)
-		*(cap->var) = val;
-	      else
-		error(0, 0, _("In GREP_COLORS=\"%s\", the \"%s\" capacity "
-			      "needs a value (\"=...\"); skipped."), p, name);
-	    }
-	  else if (val)
-	    error(0, 0, _("In GREP_COLORS=\"%s\", the \"%s\" capacity "
-			  "is boolean and cannot take a value (\"=%s\"); "
-			  "skipped."), p, name, val);
+	  {
+	    if (cap->var)
+	      {
+	        if (val)
+              *(cap->var) = val;
+	        else
+              error(0, 0, _("In GREP_COLORS=\"%s\", the \"%s\" capacity "
+			        "needs a value (\"=...\"); skipped."), p, name);
+          }
+        else if (val)
+          error(0, 0, _("In GREP_COLORS=\"%s\", the \"%s\" capacity "
+                "is boolean and cannot take a value (\"=%s\"); "
+                "skipped."), p, name, val);
+      }
 	if (cap->fct)
 	  {
 	    const char *err_str = cap->fct();
@@ -1907,6 +1914,7 @@ main (int argc, char **argv)
 
       case 'H':
 	with_filenames = 1;
+	no_filenames = 0;
 	break;
 
       case 'I':
@@ -1987,6 +1995,7 @@ main (int argc, char **argv)
 	break;
 
       case 'h':
+	with_filenames = 0;
 	no_filenames = 1;
 	break;
 
@@ -2180,10 +2189,11 @@ main (int argc, char **argv)
     {
       printf ("%s\n\n", PACKAGE_STRING);
       printf (_("\
-Copyright (C) 1988, 1992-2002, 2004, 2005  Free Software Foundation, Inc.\n"));
-      printf (_("\
-This is free software; see the source for copying conditions.  There is NO\n\
-warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"));
+Copyright (C) %s Free Software Foundation, Inc.\n\
+License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
+This is free software: you are free to change and redistribute it.\n\
+There is NO WARRANTY, to the extent permitted by law.\n"),
+        "2009");
       printf ("\n");
       exit (0);
     }
