@@ -1,9 +1,9 @@
 /* search.c - searching subroutines using dfa, kwset and regex for grep.
-   Copyright 1992, 1998, 2000 Free Software Foundation, Inc.
+   Copyright 1992, 1998, 2000, 2007, 2009 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
+   the Free Software Foundation; either version 3, or (at your option)
    any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -13,26 +13,30 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.  */
+   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
+   02110-1301, USA.  */
 
 /* Written August 1992 by Mike Haertel. */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
+
 #include <sys/types.h>
-#if defined HAVE_WCTYPE_H && defined HAVE_WCHAR_H && defined HAVE_MBRTOWC
-/* We can handle multibyte string.  */
-# define MBS_SUPPORT
+
+#include "mbsupport.h"
+#ifdef MBS_SUPPORT
+/* We can handle multibyte strings. */
 # include <wchar.h>
 # include <wctype.h>
 #endif
 
 #include "system.h"
 #include "grep.h"
-#include "regex.h"
-#include "dfa.h"
+#ifndef FGREP_PROGRAM
+# include <regex.h>
+# include "dfa.h"
+#endif
 #include "kwset.h"
 #include "error.h"
 #include "xalloc.h"
@@ -45,6 +49,26 @@
 /* For -w, we also consider _ to be word constituent.  */
 #define WCHAR(C) (ISALNUM(C) || (C) == '_')
 
+/* KWset compiled pattern.  For Ecompile and Gcompile, we compile
+   a list of strings, at least one of which is known to occur in
+   any string matching the regexp. */
+static kwset_t kwset;
+
+static void
+kwsinit (void)
+{
+  static char trans[NCHAR];
+  int i;
+
+  if (match_icase)
+    for (i = 0; i < NCHAR; ++i)
+      trans[i] = TOLOWER (i);
+
+  if (!(kwset = kwsalloc (match_icase ? trans : (char *) 0)))
+    error (2, 0, _("memory exhausted"));
+}
+
+#ifndef FGREP_PROGRAM
 /* DFA compiled regexp. */
 static struct dfa dfa;
 
@@ -60,48 +84,16 @@ static struct patterns
 struct patterns *patterns;
 size_t pcount;
 
-/* KWset compiled pattern.  For Ecompile and Gcompile, we compile
-   a list of strings, at least one of which is known to occur in
-   any string matching the regexp. */
-static kwset_t kwset;
-
-/* Number of compiled fixed strings known to exactly match the regexp.
-   If kwsexec returns < kwset_exact_matches, then we don't need to
-   call the regexp matcher at all. */
-static int kwset_exact_matches;
-
-#if defined(MBS_SUPPORT)
-static char* check_multibyte_string PARAMS ((char const *buf, size_t size));
-#endif
-static void kwsinit PARAMS ((void));
-static void kwsmusts PARAMS ((void));
-static void Gcompile PARAMS ((char const *, size_t));
-static void Ecompile PARAMS ((char const *, size_t));
-static size_t EGexecute PARAMS ((char const *, size_t, size_t *, int ));
-static void Fcompile PARAMS ((char const *, size_t));
-static size_t Fexecute PARAMS ((char const *, size_t, size_t *, int));
-static void Pcompile PARAMS ((char const *, size_t ));
-static size_t Pexecute PARAMS ((char const *, size_t, size_t *, int));
-
 void
 dfaerror (char const *mesg)
 {
   error (2, 0, mesg);
 }
 
-static void
-kwsinit (void)
-{
-  static char trans[NCHAR];
-  int i;
-
-  if (match_icase)
-    for (i = 0; i < NCHAR; ++i)
-      trans[i] = TOLOWER (i);
-
-  if (!(kwset = kwsalloc (match_icase ? trans : (char *) 0)))
-    error (2, 0, _("memory exhausted"));
-}
+/* Number of compiled fixed strings known to exactly match the regexp.
+   If kwsexec returns < kwset_exact_matches, then we don't need to
+   call the regexp matcher at all. */
+static int kwset_exact_matches;
 
 /* If the DFA turns out to have some set of fixed strings one of
    which must occur in the match, then we build a kwset matcher
@@ -140,133 +132,73 @@ kwsmusts (void)
 	error (2, 0, err);
     }
 }
+#endif /* !FGREP_PROGRAM */
 
 #ifdef MBS_SUPPORT
 /* This function allocate the array which correspond to "buf".
    Then this check multibyte string and mark on the positions which
-   are not singlebyte character nor the first byte of a multibyte
+   are not single byte character nor the first byte of a multibyte
    character.  Caller must free the array.  */
 static char*
 check_multibyte_string(char const *buf, size_t size)
 {
-  char *mb_properties = malloc(size);
+  char *mb_properties = xmalloc(size);
   mbstate_t cur_state;
+  wchar_t wc;
   int i;
+
   memset(&cur_state, 0, sizeof(mbstate_t));
   memset(mb_properties, 0, sizeof(char)*size);
+
   for (i = 0; i < size ;)
     {
       size_t mbclen;
-      mbclen = mbrlen(buf + i, size - i, &cur_state);
+      mbclen = mbrtowc(&wc, buf + i, size - i, &cur_state);
 
       if (mbclen == (size_t) -1 || mbclen == (size_t) -2 || mbclen == 0)
 	{
 	  /* An invalid sequence, or a truncated multibyte character.
-	     We treat it as a singlebyte character.  */
+	     We treat it as a single byte character.  */
 	  mbclen = 1;
 	}
+      else if (match_icase)
+	{
+	  if (iswupper((wint_t)wc))
+	    {
+	      wc = towlower((wint_t)wc);
+	      wcrtomb(buf + i, wc, &cur_state);
+	    }
+	}
       mb_properties[i] = mbclen;
       i += mbclen;
     }
 
   return mb_properties;
 }
-#endif
+#endif /* MBS_SUPPORT */
 
-static void
-Gcompile (char const *pattern, size_t size)
+#if defined(GREP_PROGRAM) || defined(EGREP_PROGRAM)
+#ifdef EGREP_PROGRAM
+COMPILE_FCT(Ecompile)
 {
-  const char *err;
-  char const *sep;
-  size_t total = size;
-  char const *motif = pattern;
-
-  re_set_syntax (RE_SYNTAX_GREP | RE_HAT_LISTS_NOT_NEWLINE);
-  dfasyntax (RE_SYNTAX_GREP | RE_HAT_LISTS_NOT_NEWLINE, match_icase, eolbyte);
-
-  /* For GNU regex compiler we have to pass the patterns separately to detect
-     errors like "[\nallo\n]\n".  The patterns here are "[", "allo" and "]"
-     GNU regex should have raise a syntax error.  The same for backref, where
-     the backref should have been local to each pattern.  */
-  do
-    {
-      size_t len;
-      sep = memchr (motif, '\n', total);
-      if (sep)
-	{
-	  len = sep - motif;
-	  sep++;
-	  total -= (len + 1);
-	}
-      else
-	{
-	  len = total;
-	  total = 0;
-	}
-
-      patterns = realloc (patterns, (pcount + 1) * sizeof (*patterns));
-      if (patterns == NULL)
-	error (2, errno, _("memory exhausted"));
-
-      patterns[pcount] = patterns0;
-
-      if ((err = re_compile_pattern (motif, len,
-				    &(patterns[pcount].regexbuf))) != 0)
-	error (2, 0, err);
-      pcount++;
-
-      motif = sep;
-    } while (sep && total != 0);
-
-  /* In the match_words and match_lines cases, we use a different pattern
-     for the DFA matcher that will quickly throw out cases that won't work.
-     Then if DFA succeeds we do some hairy stuff using the regex matcher
-     to decide whether the match should really count. */
-  if (match_words || match_lines)
-    {
-      /* In the whole-word case, we use the pattern:
-	 \(^\|[^[:alnum:]_]\)\(userpattern\)\([^[:alnum:]_]|$\).
-	 In the whole-line case, we use the pattern:
-	 ^\(userpattern\)$.  */
-
-      static char const line_beg[] = "^\\(";
-      static char const line_end[] = "\\)$";
-      static char const word_beg[] = "\\(^\\|[^[:alnum:]_]\\)\\(";
-      static char const word_end[] = "\\)\\([^[:alnum:]_]\\|$\\)";
-      char *n = malloc (sizeof word_beg - 1 + size + sizeof word_end);
-      size_t i;
-      strcpy (n, match_lines ? line_beg : word_beg);
-      i = strlen (n);
-      memcpy (n + i, pattern, size);
-      i += size;
-      strcpy (n + i, match_lines ? line_end : word_end);
-      i += strlen (n + i);
-      pattern = n;
-      size = i;
-    }
-
-  dfacomp (pattern, size, &dfa, 1);
-  kwsmusts ();
-}
-
-static void
-Ecompile (char const *pattern, size_t size)
+  reg_syntax_t syntax_bits = RE_SYNTAX_POSIX_EGREP;
+#else
+/* No __VA_ARGS__ in C89.  So we have to do it this way.  */
+static COMPILE_RET
+GEAcompile (char const *pattern, size_t size, reg_syntax_t syntax_bits)
 {
+#endif /* EGREP_PROGRAM */
   const char *err;
   const char *sep;
   size_t total = size;
   char const *motif = pattern;
 
-  if (strcmp (matcher, "awk") == 0)
-    {
-      re_set_syntax (RE_SYNTAX_AWK);
-      dfasyntax (RE_SYNTAX_AWK, match_icase, eolbyte);
-    }
-  else
-    {
-      re_set_syntax (RE_SYNTAX_POSIX_EGREP);
-      dfasyntax (RE_SYNTAX_POSIX_EGREP, match_icase, eolbyte);
-    }
+#if 0
+  if (match_icase)
+    syntax_bits |= RE_ICASE;
+#endif
+  re_set_syntax (syntax_bits);
+  dfasyntax (syntax_bits, match_icase, eolbyte);
 
   /* For GNU regex compiler we have to pass the patterns separately to detect
      errors like "[\nallo\n]\n".  The patterns here are "[", "allo" and "]"
@@ -307,66 +239,99 @@ Ecompile (char const *pattern, size_t size)
      to decide whether the match should really count. */
   if (match_words || match_lines)
     {
-      /* In the whole-word case, we use the pattern:
-	 (^|[^[:alnum:]_])(userpattern)([^[:alnum:]_]|$).
-	 In the whole-line case, we use the pattern:
-	 ^(userpattern)$.  */
+      static char const line_beg_no_bk[] = "^(";
+      static char const line_end_no_bk[] = ")$";
+      static char const word_beg_no_bk[] = "(^|[^[:alnum:]_])(";
+      static char const word_end_no_bk[] = ")([^[:alnum:]_]|$)";
+#ifdef EGREP_PROGRAM
+# define IF_BK(x, y) (y)
+      char *n = xmalloc (sizeof word_beg_no_bk - 1 + size + sizeof word_end_no_bk);
+#else
+      static char const line_beg_bk[] = "^\\(";
+      static char const line_end_bk[] = "\\)$";
+      static char const word_beg_bk[] = "\\(^\\|[^[:alnum:]_]\\)\\(";
+      static char const word_end_bk[] = "\\)\\([^[:alnum:]_]\\|$\\)";
+      int bk = !(syntax_bits & RE_NO_BK_PARENS);
+# define IF_BK(x, y) ((bk) ? (x) : (y))
+      char *n = xmalloc (sizeof word_beg_bk - 1 + size + sizeof word_end_bk);
+#endif /* EGREP_PROGRAM */
 
-      static char const line_beg[] = "^(";
-      static char const line_end[] = ")$";
-      static char const word_beg[] = "(^|[^[:alnum:]_])(";
-      static char const word_end[] = ")([^[:alnum:]_]|$)";
-      char *n = malloc (sizeof word_beg - 1 + size + sizeof word_end);
-      size_t i;
-      strcpy (n, match_lines ? line_beg : word_beg);
-      i = strlen(n);
-      memcpy (n + i, pattern, size);
-      i += size;
-      strcpy (n + i, match_lines ? line_end : word_end);
-      i += strlen (n + i);
-      pattern = n;
-      size = i;
+      strcpy (n, match_lines ? IF_BK(line_beg_bk, line_beg_no_bk)
+			     : IF_BK(word_beg_bk, word_beg_no_bk));
+      total = strlen(n);
+      memcpy (n + total, pattern, size);
+      total += size;
+      strcpy (n + total, match_lines ? IF_BK(line_end_bk, line_end_no_bk)
+				     : IF_BK(word_end_bk, word_end_no_bk));
+      total += strlen (n + total);
+      pattern = motif = n;
+      size = total;
     }
+  else
+    motif = NULL;
 
   dfacomp (pattern, size, &dfa, 1);
   kwsmusts ();
+
+  if (motif)
+    free((char *) motif);
 }
 
-static size_t
-EGexecute (char const *buf, size_t size, size_t *match_size, int exact)
+#ifndef EGREP_PROGRAM
+COMPILE_FCT(Gcompile)
 {
-  register char const *buflim, *beg, *end;
+  return GEAcompile (pattern, size,
+		     RE_SYNTAX_GREP | RE_HAT_LISTS_NOT_NEWLINE);
+}
+
+COMPILE_FCT(Acompile)
+{
+  return GEAcompile (pattern, size, RE_SYNTAX_AWK);
+}
+
+COMPILE_FCT(Ecompile)
+{
+  return GEAcompile (pattern, size, RE_SYNTAX_POSIX_EGREP);
+}
+#endif /* !EGREP_PROGRAM */
+
+EXECUTE_FCT(EGexecute)
+{
+  register char const *buflim, *beg, *end, *match, *best_match;
   char eol = eolbyte;
-  int backref, start, len;
+  int backref, start, len, best_len;
   struct kwsmatch kwsm;
-  size_t i;
+  size_t i, ret_val;
 #ifdef MBS_SUPPORT
   char *mb_properties = NULL;
-#endif /* MBS_SUPPORT */
-
-#ifdef MBS_SUPPORT
-  if (MB_CUR_MAX > 1 && kwset)
-    mb_properties = check_multibyte_string(buf, size);
+  if (MB_CUR_MAX > 1)
+    {
+      if (match_icase)
+        {
+          char *case_buf = xmalloc(size);
+          memcpy(case_buf, buf, size);
+	  if (start_ptr)
+	    start_ptr = case_buf + (start_ptr - buf);
+          buf = case_buf;
+        }
+      if (kwset)
+        mb_properties = check_multibyte_string(buf, size);
+    }
 #endif /* MBS_SUPPORT */
 
   buflim = buf + size;
 
   for (beg = end = buf; end < buflim; beg = end)
     {
-      if (!exact)
+      if (!start_ptr)
 	{
+	  /* We don't care about an exact match.  */
 	  if (kwset)
 	    {
 	      /* Find a possible match using the KWset matcher. */
 	      size_t offset = kwsexec (kwset, beg, buflim - beg, &kwsm);
 	      if (offset == (size_t) -1)
-		{
-#ifdef MBS_SUPPORT
-		  if (MB_CUR_MAX > 1)
-		    free(mb_properties);
-#endif
-		  return (size_t)-1;
-		}
+		goto failure;
 	      beg += offset;
 	      /* Narrow down to the line containing the candidate, and
 		 run it through DFA. */
@@ -401,26 +366,38 @@ EGexecute (char const *buf, size_t size, size_t *match_size, int exact)
 	    goto success;
 	}
       else
-	end = beg + size;
+	{
+	  /* We are looking for the leftmost (then longest) exact match.
+	     We will go through the outer loop only once.  */
+	  beg = start_ptr;
+	  end = buflim;
+	}
 
       /* If we've made it to this point, this means DFA has seen
 	 a probable match, and we need to run it through Regex. */
+      best_match = end;
+      best_len = 0;
       for (i = 0; i < pcount; i++)
 	{
 	  patterns[i].regexbuf.not_eol = 0;
-	  if (0 <= (start = re_search (&(patterns[i].regexbuf), beg,
-				       end - beg - 1, 0,
-				       end - beg - 1, &(patterns[i].regs))))
+	  if (0 <= (start = re_search (&(patterns[i].regexbuf),
+				       buf, end - buf - 1,
+				       beg - buf, end - beg - 1,
+				       &(patterns[i].regs))))
 	    {
 	      len = patterns[i].regs.end[0] - start;
-	      if (exact)
-		{
-		  *match_size = len;
-		  return start;
-		}
+	      match = buf + start;
+	      if (match > best_match)
+		continue;
+	      if (start_ptr && !match_words)
+		goto assess_pattern_match;
 	      if ((!match_lines && !match_words)
 		  || (match_lines && len == end - beg - 1))
-		goto success;
+		{
+		  match = beg;
+		  len = end - beg;
+		  goto assess_pattern_match;
+		}
 	      /* If -w, check if the match aligns with word boundaries.
 		 We do this iteratively because:
 		 (a) the line may contain more than one occurence of the
@@ -429,55 +406,90 @@ EGexecute (char const *buf, size_t size, size_t *match_size, int exact)
 		 given point, and we may need to consider a shorter one to
 		 find a word boundary.  */
 	      if (match_words)
-		while (start >= 0)
+		while (match <= best_match)
 		  {
-		    if ((start == 0 || !WCHAR ((unsigned char) beg[start - 1]))
+		    if ((match == buf || !WCHAR ((unsigned char) match[-1]))
 			&& (len == end - beg - 1
-			    || !WCHAR ((unsigned char) beg[start + len])))
-		      goto success;
+			    || !WCHAR ((unsigned char) match[len])))
+		      goto assess_pattern_match;
 		    if (len > 0)
 		      {
 			/* Try a shorter length anchored at the same place. */
 			--len;
 			patterns[i].regexbuf.not_eol = 1;
-			len = re_match (&(patterns[i].regexbuf), beg,
-					start + len, start,
+			len = re_match (&(patterns[i].regexbuf),
+					buf, match + len - beg, match - buf,
 					&(patterns[i].regs));
 		      }
 		    if (len <= 0)
 		      {
 			/* Try looking further on. */
-			if (start == end - beg - 1)
+			if (match == end - 1)
 			  break;
-			++start;
+			match++;
 			patterns[i].regexbuf.not_eol = 0;
-			start = re_search (&(patterns[i].regexbuf), beg,
-					   end - beg - 1,
-					   start, end - beg - 1 - start,
+			start = re_search (&(patterns[i].regexbuf),
+					   buf, end - buf - 1,
+					   match - buf, end - match - 1,
 					   &(patterns[i].regs));
+			if (start < 0)
+			  break;
 			len = patterns[i].regs.end[0] - start;
+			match = buf + start;
 		      }
-		  }
-	    }
+		  } /* while (match <= best_match) */
+	      continue;
+	    assess_pattern_match:
+	      if (!start_ptr)
+		{
+		  /* Good enough for a non-exact match.
+		     No need to look at further patterns, if any.  */
+		  beg = match;
+		  goto success_in_len;
+		}
+	      if (match < best_match || (match == best_match && len > best_len))
+		{
+		  /* Best exact match:  leftmost, then longest.  */
+		  best_match = match;
+		  best_len = len;
+		}
+	    } /* if re_search >= 0 */
 	} /* for Regex patterns.  */
+	if (best_match < end)
+	  {
+	    /* We have found an exact match.  We were just
+	       waiting for the best one (leftmost then longest).  */
+	    beg = best_match;
+	    len = best_len;
+	    goto success_in_len;
+	  }
     } /* for (beg = end ..) */
-#ifdef MBS_SUPPORT
-  if (MB_CUR_MAX > 1 && mb_properties)
-    free (mb_properties);
-#endif /* MBS_SUPPORT */
-  return (size_t) -1;
+
+ failure:
+  ret_val = -1;
+  goto out;
 
  success:
+  len = end - beg;
+ success_in_len:
+  *match_size = len;
+  ret_val = beg - buf;
+ out:
 #ifdef MBS_SUPPORT
-  if (MB_CUR_MAX > 1 && mb_properties)
-    free (mb_properties);
+  if (MB_CUR_MAX > 1)
+    {
+      if (match_icase)
+        free((char*)buf);
+      if (mb_properties)
+        free(mb_properties);
+    }
 #endif /* MBS_SUPPORT */
-  *match_size = end - beg;
-  return beg - buf;
+  return ret_val;
 }
+#endif /* defined(GREP_PROGRAM) || defined(EGREP_PROGRAM) */
 
-static void
-Fcompile (char const *pattern, size_t size)
+#if defined(GREP_PROGRAM) || defined(FGREP_PROGRAM)
+COMPILE_FCT(Fcompile)
 {
   char const *beg, *lim, *err;
 
@@ -499,45 +511,42 @@ Fcompile (char const *pattern, size_t size)
     error (2, 0, err);
 }
 
-static size_t
-Fexecute (char const *buf, size_t size, size_t *match_size, int exact)
+EXECUTE_FCT(Fexecute)
 {
   register char const *beg, *try, *end;
   register size_t len;
   char eol = eolbyte;
   struct kwsmatch kwsmatch;
+  size_t ret_val;
 #ifdef MBS_SUPPORT
-  char *mb_properties;
+  char *mb_properties = NULL;
   if (MB_CUR_MAX > 1)
-    mb_properties = check_multibyte_string (buf, size);
+    {
+      if (match_icase)
+        {
+          char *case_buf = xmalloc(size);
+          memcpy(case_buf, buf, size);
+	  if (start_ptr)
+	    start_ptr = case_buf + (start_ptr - buf);
+          buf = case_buf;
+        }
+      mb_properties = check_multibyte_string(buf, size);
+    }
 #endif /* MBS_SUPPORT */
 
-  for (beg = buf; beg <= buf + size; ++beg)
+  for (beg = start_ptr ? start_ptr : buf; beg <= buf + size; beg++)
     {
       size_t offset = kwsexec (kwset, beg, buf + size - beg, &kwsmatch);
       if (offset == (size_t) -1)
-	{
-#ifdef MBS_SUPPORT
-	  if (MB_CUR_MAX > 1)
-	    free(mb_properties);
-#endif /* MBS_SUPPORT */
-	  return offset;
-	}
+	goto failure;
 #ifdef MBS_SUPPORT
       if (MB_CUR_MAX > 1 && mb_properties[offset+beg-buf] == 0)
 	continue; /* It is a part of multibyte character.  */
 #endif /* MBS_SUPPORT */
       beg += offset;
       len = kwsmatch.size[0];
-      if (exact)
-	{
-	  *match_size = len;
-#ifdef MBS_SUPPORT
-	  if (MB_CUR_MAX > 1)
-	    free (mb_properties);
-#endif /* MBS_SUPPORT */
-	  return beg - buf;
-	}
+      if (start_ptr && !match_words)
+	goto success_in_beg_and_len;
       if (match_lines)
 	{
 	  if (beg > buf && beg[-1] != eol)
@@ -555,42 +564,47 @@ Fexecute (char const *buf, size_t size, size_t *match_size, int exact)
 	      {
 		offset = kwsexec (kwset, beg, --len, &kwsmatch);
 		if (offset == (size_t) -1)
-		  {
-#ifdef MBS_SUPPORT
-		    if (MB_CUR_MAX > 1)
-		      free (mb_properties);
-#endif /* MBS_SUPPORT */
-		    return offset;
-		  }
+		  break;
 		try = beg + offset;
 		len = kwsmatch.size[0];
 	      }
-	    else
+	    else if (!start_ptr)
 	      goto success;
-	  }
+	    else
+	      goto success_in_beg_and_len;
+	  } /* for (try) */
       else
 	goto success;
-    }
+    } /* for (beg in buf) */
 
-#ifdef MBS_SUPPORT
-  if (MB_CUR_MAX > 1)
-    free (mb_properties);
-#endif /* MBS_SUPPORT */
-  return -1;
+ failure:
+  ret_val = -1;
+  goto out;
 
  success:
   end = memchr (beg + len, eol, (buf + size) - (beg + len));
   end++;
   while (buf < beg && beg[-1] != eol)
     --beg;
-  *match_size = end - beg;
+  len = end - beg;
+ success_in_beg_and_len:
+  *match_size = len;
+  ret_val = beg - buf;
+ out:
 #ifdef MBS_SUPPORT
   if (MB_CUR_MAX > 1)
-    free (mb_properties);
+    {
+      if (match_icase)
+        free((char*)buf);
+      if (mb_properties)
+        free(mb_properties);
+    }
 #endif /* MBS_SUPPORT */
-  return beg - buf;
+  return ret_val;
 }
+#endif /* defined(GREP_PROGRAM) || defined(FGREP_PROGRAM) */
 
+#ifdef GREP_PROGRAM
 #if HAVE_LIBPCRE
 /* Compiled internal form of a Perl regular expression.  */
 static pcre *cre;
@@ -599,11 +613,10 @@ static pcre *cre;
 static pcre_extra *extra;
 #endif
 
-static void
-Pcompile (char const *pattern, size_t size)
+COMPILE_FCT(Pcompile)
 {
 #if !HAVE_LIBPCRE
-  error (2, 0, _("The -P option is not supported"));
+  error (2, 0, "%s", _("Support for the -P option is not compiled into this --disable-perl-regexp binary"));
 #else
   int e;
   char const *ep;
@@ -614,9 +627,11 @@ Pcompile (char const *pattern, size_t size)
   char const *p;
   char const *pnul;
 
-  /* FIXME: Remove this restriction.  */
+  /* FIXME: Remove these restrictions.  */
   if (eolbyte != '\n')
     error (2, 0, _("The -P and -z options cannot be combined"));
+  if (memchr(pattern, '\n', size))
+    error (2, 0, _("The -P option only supports a single pattern"));
 
   *n = '\0';
   if (match_lines)
@@ -664,8 +679,7 @@ Pcompile (char const *pattern, size_t size)
 #endif
 }
 
-static size_t
-Pexecute (char const *buf, size_t size, size_t *match_size, int exact)
+EXECUTE_FCT(Pexecute)
 {
 #if !HAVE_LIBPCRE
   abort ();
@@ -675,7 +689,8 @@ Pexecute (char const *buf, size_t size, size_t *match_size, int exact)
      is just for performance improvement in pcre_exec.  */
   int sub[300];
 
-  int e = pcre_exec (cre, extra, buf, size, 0, 0,
+  int e = pcre_exec (cre, extra, buf, size,
+		     start_ptr ? (start_ptr - buf) : 0, 0,
 		     sub, sizeof sub / sizeof *sub);
 
   if (e <= 0)
@@ -699,10 +714,16 @@ Pexecute (char const *buf, size_t size, size_t *match_size, int exact)
       char const *end = buf + sub[1];
       char const *buflim = buf + size;
       char eol = eolbyte;
-      if (!exact)
+      if (!start_ptr)
 	{
-	  end = memchr (end, eol, buflim - end);
-	  end++;
+	  /* FIXME: The case when '\n' is not found indicates a bug:
+	     Since grep is line oriented, the match should never contain
+	     a newline, so there _must_ be a newline following.
+	   */
+	  if (!(end = memchr (end, eol, buflim - end)))
+	    end = buflim;
+	  else
+	    end++;
 	  while (buf < beg && beg[-1] != eol)
 	    --beg;
 	}
@@ -715,10 +736,11 @@ Pexecute (char const *buf, size_t size, size_t *match_size, int exact)
 
 struct matcher const matchers[] = {
   { "default", Gcompile, EGexecute },
-  { "grep", Gcompile, EGexecute },
-  { "egrep", Ecompile, EGexecute },
-  { "awk", Ecompile, EGexecute },
-  { "fgrep", Fcompile, Fexecute },
-  { "perl", Pcompile, Pexecute },
+  { "grep",    Gcompile, EGexecute },
+  { "egrep",   Ecompile, EGexecute },
+  { "awk",     Acompile, EGexecute },
+  { "fgrep",   Fcompile, Fexecute },
+  { "perl",    Pcompile, Pexecute },
   { "", 0, 0 },
 };
+#endif /* GREP_PROGRAM */
