@@ -1,7 +1,11 @@
-#!/bin/sh
-# Test that backrefs are local to regex.
-#
+#! /bin/sh
+# Test for backreferences and other things.
+# 
+# Copyright (C) 2001, 2006, 2009 Free Software Foundation, Inc.
 #
+# Copying and distribution of this file, with or without modification,
+# are permitted in any medium without royalty provided the copyright
+# notice and this notice are preserved.
 
 : ${srcdir=.}
 
@@ -10,7 +14,7 @@ failures=0
 # checking for a palindrome
 echo "radar" | ${GREP} -e '\(.\)\(.\).\2\1' > /dev/null 2>&1
 if test $? -ne 0 ; then
-        echo "backref: palindrome, test \#1 failed"
+        echo "Backref: palindrome, test #1 failed"
         failures=1
 fi
 
@@ -18,21 +22,21 @@ fi
 # For now, remove the ``?'' in the last parentheses, so that new glibc can do it.  --Stepan
 echo "civic" | ${GREP} -E -e '^(.?)(.?)(.?)(.?)(.?)(.?)(.?)(.?)(.).?\9\8\7\6\5\4\3\2\1$' > /dev/null 2>&1
 if test $? -ne 0 ; then
-        echo "Options: Bond, test \#2 failed"
+        echo "Options: Bond, test #2 failed"
         failures=1
 fi
 
 # backref are local should be error
 echo "123" | ${GREP} -e 'a\(.\)' -e 'b\1' > /dev/null 2>&1
 if test $? -ne 2 ; then
-	echo "Options: Backref not local, test \#3 failed"
+	echo "Backref: Backref not local, test #3 failed"
 	failures=1
 fi
 
-# Pattern should faile
+# Pattern should fail
 echo "123" | ${GREP} -e '[' -e ']' > /dev/null 2>&1
 if test $? -ne 2 ; then
-	echo "Options: Compiled not local, test \#3 failed"
+	echo "Backref: Compiled not local, test #4 failed"
 	failures=1
 fi
 
